﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using System.IO;
using System.IO.Ports;
using System.ComponentModel;

namespace rae2_test
{
    /// <summary>
    /// MainWindow.xaml の相互作用ロジック
    /// </summary>
    public partial class MainWindow : Window, INotifyPropertyChanged
    {
        public event PropertyChangedEventHandler PropertyChanged;
        private void OnPropertyChanged(string propertyName)
        {
            if (this.PropertyChanged != null)
            {
                PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
            }
        }

        private delegate void LockButtonsDelegate();
        private delegate void UnlockButtonsDelegate();

        class DoseRae2Dispatcher
        {
            DoseRae2Communicator doseRae2 = null;
            private Object queueLock = new Object();
            private volatile bool doseRae2dispatchStop = false;

            private class QueueItem
            {
                public enum RequestType
                {
                    ALLLOGSAVE = 0
                }

                public RequestType Type;
                public object Argument;

                public QueueItem(RequestType type, object argument)
                {
                    this.Type = type;
                    this.Argument = argument;
                }
            }

            private volatile Queue<QueueItem> queueList = new Queue<QueueItem>();
            private bool processing = false;

            AddOutputWindowDelegate AddOutputWindow = null;
            LockButtonsDelegate LockButtons = null;
            UnlockButtonsDelegate UnlockButtons = null;

            public DoseRae2Dispatcher(
                DoseRae2Communicator communicator,
                AddOutputWindowDelegate addOutputWindowDelegate,
                LockButtonsDelegate lockButtonsDelegate,
                UnlockButtonsDelegate unlockButtonsDelegate)
            {
                doseRae2 = communicator;
                AddOutputWindow = addOutputWindowDelegate;
                LockButtons = lockButtonsDelegate;
                UnlockButtons = unlockButtonsDelegate;
            }

            public void Stop()
            {
                doseRae2dispatchStop = true;
                if (doseRae2 != null)
                {
                    doseRae2.Stop();
                }
            }

            public void dispatch()
            {
                while (!doseRae2dispatchStop)
                {
                    QueueItem item = null;
                    lock (queueLock)
                    {
                        if (queueList.Count != 0)
                        {
                            item = queueList.Dequeue();
                        }
                    }

                    if (item != null)
                    {
                        switch (item.Type)
                        {
                            case QueueItem.RequestType.ALLLOGSAVE:
                                allLogSaveProcess(item.Argument);
                                break;
                        }

                        lock (queueLock)
                        {
                            // ロックの都合で、QueueIsEmptyは使えない
                            if (queueList.Count == 0)
                            {
                                this.UnlockButtons();
                            }
                        }
                    }

                    Thread.Sleep(1);
                }
            }

            public bool QueueIsEmpty()
            {
                lock (queueLock)
                {
                    return queueList.Count == 0 && !processing;
                }
            }

            private void AddQueue(QueueItem item)
            {
                lock (queueLock)
                {
                    queueList.Enqueue(item);
                    this.LockButtons();
                }
            }

            public void allLogSaveProcess(object arg)
            {
                string filename = (string)arg;

                if (!doseRae2.CommandConnectDevice())
                {
                    this.AddOutputWindow("デバイスとの接続に失敗しました");
                    return;
                }

                List<LogData> log = doseRae2.CommandGetAllData();
                if (log == null)
                {
                    return;
                }

                StreamWriter writer = new StreamWriter(@"test.csv", true, Encoding.GetEncoding("Shift_JIS"));
                foreach (LogData data in log)
                {
                    string line = string.Format("{0:d4}/{1:d2}/{2:d2} {3:d2}:{4:d2}:{5:d2},{6:d}",
                        data.time.Year, data.time.Month, data.time.Day,
                        data.time.Hour, data.time.Minute, data.time.Second,
                        data.value);
                    writer.WriteLine(line);
                }
                writer.Close();
            }

            public void AllLogSaveAsync(string filename)
            {
                AddQueue(new QueueItem(QueueItem.RequestType.ALLLOGSAVE, filename));
            }

            public bool PortOpenSync(string comport)
            {
                while (!QueueIsEmpty())
                {
                    Thread.Sleep(1);
                }

                return doseRae2.PortOpen(comport);
            }

            public bool PortCloseSync()
            {
                while (!QueueIsEmpty())
                {
                    Thread.Sleep(1);
                }

                return doseRae2.PortClose();
            }

            public List<LogData> CommandGetAllDataSync()
            {
                while (!QueueIsEmpty())
                {
                    Thread.Sleep(1);
                }

                if (!doseRae2.CommandConnectDevice())
                {
                    this.AddOutputWindow("デバイスとの接続に失敗しました");
                    return null;
                }

                return doseRae2.CommandGetAllData();
            }

            public void CommandClearLogSync()
            {
                while (!QueueIsEmpty())
                {
                    Thread.Sleep(1);
                }

                if (!doseRae2.CommandConnectDevice())
                {
                    this.AddOutputWindow("デバイスとの接続に失敗しました");
                    return;
                }

                doseRae2.CommandClearLog();
            }

            public void CommandClearDSync()
            {
                while (!QueueIsEmpty())
                {
                    Thread.Sleep(1);
                }

                if (!doseRae2.CommandConnectDevice())
                {
                    this.AddOutputWindow("デバイスとの接続に失敗しました");
                    return;
                }

                doseRae2.CommandClearD();
            }
        }


        private DoseRae2Dispatcher dispatcher = null;
        private Thread dispatcherThread = null;

        public MainWindow()
        {
            PortList = new List<string>();

            InitializeComponent();

            if (this.UpdatePortList() > 0)
            {
                this.comboBox1.SelectedIndex = 0;
            }

            DoseRae2Communicator comm = new DoseRae2Communicator(new AddOutputWindowDelegate(this.AddOutputWindow));

            dispatcher = new DoseRae2Dispatcher(
                comm,
                new AddOutputWindowDelegate(this.AddOutputWindow),
                new LockButtonsDelegate(this.LockButtons),
                new UnlockButtonsDelegate(this.UnlockButtons));
            dispatcherThread = new Thread(dispatcher.dispatch);
            dispatcherThread.Start();

            this.AddOutputWindow("開始しました。");
        }

        ~MainWindow()
        {
            dispatcher.Stop();
            dispatcherThread.Join();
        }

        private void window1_Closing(object sender, CancelEventArgs e)
        {
            if (dispatcher != null)
            {
                dispatcher.Stop();
            }
        }

        private bool buttonEnable = true;
        public bool ButtonEnable
        {
            get
            {
                return buttonEnable;
            }
            set
            {
                if (buttonEnable != value)
                {
                    buttonEnable = value;
                    this.OnPropertyChanged("ButtonEnable");
                }
            }
        }

        public void LockButtons()
        {
            this.ButtonEnable = false;
        }

        public void UnlockButtons()
        {
            this.ButtonEnable = true;
        }

        public List<string> PortList{get;set;}
        private int UpdatePortList()
        {
            string[] ports = SerialPort.GetPortNames();
            foreach (string portname in ports)
            {
                PortList.Add(portname);
            }
            return PortList.Count;
        }

        private void button1_Click(object sender, RoutedEventArgs e)
        {
            //if (port == null)
            {
                // open port
                string comport = (string)this.comboBox1.SelectedItem;
                if (dispatcher.PortOpenSync(comport))
                {
                    //this.button1.Content = "Disconnect";
                    this.button1.IsEnabled = false;
                }
            }
            if (false) // else
            {
                // close port
                if (dispatcher.PortCloseSync())
                {
                    this.button1.Content = "Connect";
                }
            }
        }

        private void button2_Click(object sender, RoutedEventArgs e)
        {
            dispatcher.AllLogSaveAsync(@"test.csv");
        }

        private void button3_Click(object sender, RoutedEventArgs e)
        {
            dispatcher.CommandClearLogSync();
        }

        private void button4_Click(object sender, RoutedEventArgs e)
        {
            dispatcher.CommandClearDSync();
        }

        /* *******************************************************************
         * 履歴バッファ
         */
        #region logbuffer
        private Object messageLock = new Object();

        const int MaxLogLine = 100;

        public string statusMessage = "";
        public string StatusMessage
        {
            get
            {
                return statusMessage;
            }
            set
            {
                if (statusMessage != value)
                {
                    statusMessage = value;
                    this.OnPropertyChanged("StatusMessage");
                }
            }
        }

        public Queue<string> outputWindowBuffer = new Queue<string>();
        public string OutputWindowBuffer
        {
            get
            {
                string allMessage = "";
                foreach (string s in this.outputWindowBuffer)
                {
                    allMessage += s;
                }
                return allMessage;
            }
        }

        public void AddOutputWindow(string message)
        {
            DateTime tm = new DateTime(DateTime.Now.Ticks);

            message =
                string.Format(
                "{0:d4}/{1:d2}/{2:d2} {3:d2}:{4:d2}:{5:d2} ",
                tm.Year, tm.Month, tm.Day, tm.Hour, tm.Minute, tm.Second)
                + message
                + "\n";

            lock (messageLock)
            {
                this.StatusMessage = message;

                if (this.outputWindowBuffer.Count > MaxLogLine)
                {
                    this.outputWindowBuffer.Dequeue();
                }
                this.outputWindowBuffer.Enqueue(message);
                this.OnPropertyChanged("OutputWindowBuffer");
            }
        }
        #endregion

        /* *******************************************************************
         * ウインドウサイズ関連
         */
        #region windowsize
        private void Window_SizeChanged(object sender, SizeChangedEventArgs e)
        {
            Window w = (Window)sender;

            double h = this.window1.Height;
            h -= 40 + 50 + 24;

            LogGrid.Height = h;
        }

        private double windowHeight = 24;
        public double WindowHeight
        {
            get
            {
                return windowHeight;
            }
            set
            {
                if (windowHeight != value)
                {
                    windowHeight = value;
                    this.OnPropertyChanged("WindowHeight");
                }
            }
        }

        private void expander1_Collapsed(object sender, RoutedEventArgs e)
        {
            window1.MinHeight = 140;
            window1.MaxHeight = 140;

            window1.Height = 140;
            WindowHeight = 140;
        }

        private void expander1_Expanded(object sender, RoutedEventArgs e)
        {
            window1.MinHeight = 140;
            window1.MaxHeight = 1000;

            window1.Height = 400;
            WindowHeight = 400;
        }
        #endregion
    }
}
